import re
import cffi

import numpy as np

from numba.cuda.testing import (
    skip_if_curand_kernel_missing,
    skip_on_cudasim,
    test_data_dir,
    unittest,
    CUDATestCase,
)
from numba import cuda, jit, float32, int32, types
from numba.core.errors import TypingError
from numba.tests.support import skip_unless_cffi
from types import ModuleType


class TestDeviceFunc(CUDATestCase):
    def test_use_add2f(self):
        @cuda.jit("float32(float32, float32)", device=True)
        def add2f(a, b):
            return a + b

        def use_add2f(ary):
            i = cuda.grid(1)
            ary[i] = add2f(ary[i], ary[i])

        compiled = cuda.jit("void(float32[:])")(use_add2f)

        nelem = 10
        ary = np.arange(nelem, dtype=np.float32)
        exp = ary + ary
        compiled[1, nelem](ary)

        self.assertTrue(np.all(ary == exp), (ary, exp))

    def test_indirect_add2f(self):
        @cuda.jit("float32(float32, float32)", device=True)
        def add2f(a, b):
            return a + b

        @cuda.jit("float32(float32, float32)", device=True)
        def indirect(a, b):
            return add2f(a, b)

        def indirect_add2f(ary):
            i = cuda.grid(1)
            ary[i] = indirect(ary[i], ary[i])

        compiled = cuda.jit("void(float32[:])")(indirect_add2f)

        nelem = 10
        ary = np.arange(nelem, dtype=np.float32)
        exp = ary + ary
        compiled[1, nelem](ary)

        self.assertTrue(np.all(ary == exp), (ary, exp))

    def _check_cpu_dispatcher(self, add):
        @cuda.jit
        def add_kernel(ary):
            i = cuda.grid(1)
            ary[i] = add(ary[i], 1)

        ary = np.arange(10)
        expect = ary + 1
        add_kernel[1, ary.size](ary)
        np.testing.assert_equal(expect, ary)

    def test_cpu_dispatcher(self):
        # Test correct usage
        @jit
        def add(a, b):
            return a + b

        self._check_cpu_dispatcher(add)

    @skip_on_cudasim("not supported in cudasim")
    def test_cpu_dispatcher_invalid(self):
        # Test invalid usage
        # Explicit signature disables compilation, which also disable
        # compiling on CUDA.
        @jit("(i4, i4)")
        def add(a, b):
            return a + b

        # Check that the right error message is provided.
        with self.assertRaises(TypingError) as raises:
            self._check_cpu_dispatcher(add)
        msg = "Untyped global name 'add':.*using cpu function on device"
        expected = re.compile(msg)
        self.assertTrue(expected.search(str(raises.exception)) is not None)

    def test_cpu_dispatcher_other_module(self):
        @jit
        def add(a, b):
            return a + b

        mymod = ModuleType(name="mymod")
        mymod.add = add
        del add

        @cuda.jit
        def add_kernel(ary):
            i = cuda.grid(1)
            ary[i] = mymod.add(ary[i], 1)

        ary = np.arange(10)
        expect = ary + 1
        add_kernel[1, ary.size](ary)
        np.testing.assert_equal(expect, ary)

    @skip_on_cudasim("not supported in cudasim")
    def test_inspect_llvm(self):
        @cuda.jit(device=True)
        def foo(x, y):
            return x + y

        args = (int32, int32)
        cres = foo.compile_device(args)

        fname = cres.fndesc.mangled_name
        # Verify that the function name has "foo" in it as in the python name
        self.assertIn("foo", fname)

        llvm = foo.inspect_llvm(args)
        # Check that the compiled function name is in the LLVM.
        self.assertIn(fname, llvm)

    @skip_on_cudasim("not supported in cudasim")
    def test_inspect_asm(self):
        @cuda.jit(device=True)
        def foo(x, y):
            return x + y

        args = (int32, int32)
        cres = foo.compile_device(args)

        fname = cres.fndesc.mangled_name
        # Verify that the function name has "foo" in it as in the python name
        self.assertIn("foo", fname)

        ptx = foo.inspect_asm(args)
        # Check that the compiled function name is in the PTX
        self.assertIn(fname, ptx)

    @skip_on_cudasim("not supported in cudasim")
    def test_inspect_sass_disallowed(self):
        @cuda.jit(device=True)
        def foo(x, y):
            return x + y

        with self.assertRaises(RuntimeError) as raises:
            foo.inspect_sass((int32, int32))

        self.assertIn(
            "Cannot inspect SASS of a device function", str(raises.exception)
        )

    @skip_on_cudasim("cudasim will allow calling any function")
    def test_device_func_as_kernel_disallowed(self):
        @cuda.jit(device=True)
        def f():
            pass

        with self.assertRaises(RuntimeError) as raises:
            f[1, 1]()

        self.assertIn(
            "Cannot compile a device function as a kernel",
            str(raises.exception),
        )

    @skip_on_cudasim("cudasim ignores casting by jit decorator signature")
    def test_device_casting(self):
        # Ensure that casts to the correct type are forced when calling a
        # device function with a signature. This test ensures that:
        #
        # - We don't compile a new specialization of rgba for float32 when we
        #   shouldn't
        # - We insert a cast when calling rgba, as opposed to failing to type.

        @cuda.jit("int32(int32, int32, int32, int32)", device=True)
        def rgba(r, g, b, a):
            return (
                ((r & 0xFF) << 16)
                | ((g & 0xFF) << 8)
                | ((b & 0xFF) << 0)
                | ((a & 0xFF) << 24)
            )

        @cuda.jit
        def rgba_caller(x, channels):
            x[0] = rgba(channels[0], channels[1], channels[2], channels[3])

        x = cuda.device_array(1, dtype=np.int32)
        channels = cuda.to_device(
            np.asarray([1.0, 2.0, 3.0, 4.0], dtype=np.float32)
        )

        rgba_caller[1, 1](x, channels)

        self.assertEqual(0x04010203, x[0])


times2_cu = cuda.CUSource("""
extern "C" __device__
int times2(int *out, int a)
{
  *out = a * 2;
  return 0;
}
""")

times3_cu = cuda.CUSource("""
extern "C" __device__
int times3(int *out, int a)
{
  *out = a * 3;
  return 0;
}
""")

times4_cu = cuda.CUSource("""
extern "C" __device__
int times2(int *out, int a);

extern "C" __device__
int times4(int *out, int a)
{
  int tmp;
  times2(&tmp, a);
  *out = tmp * 2;
  return 0;
}
""")

jitlink_user_cu = cuda.CUSource("""
extern "C" __device__
int array_mutator(void *out, int *a);

extern "C" __device__
int use_array_mutator(void *out, int *a) {
  array_mutator(out, a);
  return 0;
}
""")

rng_cu = cuda.CUSource("""
#include <curand_kernel.h>

extern "C" __device__
int random_number(unsigned int *out, unsigned long long seed)
{
  // Initialize state
  curandStateXORWOW_t state;
  unsigned long long sequence = 1;
  unsigned long long offset = 0;
  curand_init(seed, sequence, offset, &state);

  // Generate one random number
  *out = curand(&state);

  // Report no exception
  return 0;
}""")


@skip_on_cudasim("External functions unsupported in the simulator")
class TestDeclareDevice(CUDATestCase):
    def check_api(self, decl):
        self.assertEqual(decl.name, "f1")
        self.assertEqual(decl.sig.args, (float32[:],))
        self.assertEqual(decl.sig.return_type, int32)

    def test_declare_device_signature(self):
        f1 = cuda.declare_device("f1", int32(float32[:]))
        self.check_api(f1)

    def test_declare_device_string(self):
        f1 = cuda.declare_device("f1", "int32(float32[:])")
        self.check_api(f1)

    def test_bad_declare_device_tuple(self):
        with self.assertRaisesRegex(TypeError, "Return type"):
            cuda.declare_device("f1", (float32[:],))

    def test_bad_declare_device_string(self):
        with self.assertRaisesRegex(TypeError, "Return type"):
            cuda.declare_device("f1", "(float32[:],)")

    def test_link_cu_source(self):
        times2 = cuda.declare_device("times2", "int32(int32)", link=times2_cu)

        @cuda.jit
        def kernel(r, x):
            i = cuda.grid(1)
            if i < len(r):
                r[i] = times2(x[i])

        x = np.arange(10, dtype=np.int32)
        r = np.empty_like(x)

        kernel[1, 32](r, x)

        np.testing.assert_equal(r, x * 2)

    def _test_link_multiple_sources(self, link_type):
        link = link_type([times2_cu, times4_cu])
        times4 = cuda.declare_device("times4", "int32(int32)", link=link)

        @cuda.jit
        def kernel(r, x):
            i = cuda.grid(1)
            if i < len(r):
                r[i] = times4(x[i])

        x = np.arange(10, dtype=np.int32)
        r = np.empty_like(x)

        kernel[1, 32](r, x)

        np.testing.assert_equal(r, x * 4)

    def test_link_multiple_sources_set(self):
        self._test_link_multiple_sources(set)

    def test_link_multiple_sources_tuple(self):
        self._test_link_multiple_sources(tuple)

    def test_link_multiple_sources_list(self):
        self._test_link_multiple_sources(list)

    @skip_unless_cffi
    def test_link_sources_in_memory_and_on_disk(self):
        jitlink_cu = str(test_data_dir / "jitlink.cu")
        link = [jitlink_cu, jitlink_user_cu]
        sig = types.void(types.CPointer(types.int32))
        ext_fn = cuda.declare_device("use_array_mutator", sig, link=link)

        ffi = cffi.FFI()

        @cuda.jit
        def kernel(x):
            ptr = ffi.from_buffer(x)
            ext_fn(ptr)

        x = np.arange(2, dtype=np.int32)
        kernel[1, 1](x)

        expected = np.ones(2, dtype=np.int32)
        np.testing.assert_equal(x, expected)

    @skip_if_curand_kernel_missing
    def test_include_cuda_header(self):
        sig = types.int32(types.uint64)
        link = [rng_cu]
        random_number = cuda.declare_device("random_number", sig, link=link)

        @cuda.jit
        def kernel(x, seed):
            x[0] = random_number(seed)

        x = np.zeros(1, dtype=np.uint32)
        kernel[1, 1](x, 1)
        np.testing.assert_equal(x[0], 323845807)

    def test_declared_in_called_function(self):
        times2 = cuda.declare_device("times2", "int32(int32)", link=times2_cu)

        @cuda.jit
        def device_func(x):
            return times2(x)

        @cuda.jit
        def kernel(r, x):
            i = cuda.grid(1)
            if i < len(r):
                r[i] = device_func(x[i])

        x = np.arange(10, dtype=np.int32)
        r = np.empty_like(x)

        kernel[1, 32](r, x)

        np.testing.assert_equal(r, x * 2)

    def test_declared_in_called_function_twice(self):
        times2 = cuda.declare_device("times2", "int32(int32)", link=times2_cu)

        @cuda.jit
        def device_func_1(x):
            return times2(x)

        @cuda.jit
        def device_func_2(x):
            return device_func_1(x)

        @cuda.jit
        def kernel(r, x):
            i = cuda.grid(1)
            if i < len(r):
                r[i] = device_func_2(x[i])

        x = np.arange(10, dtype=np.int32)
        r = np.empty_like(x)

        kernel[1, 32](r, x)

        np.testing.assert_equal(r, x * 2)

    def test_declared_in_called_function_two_calls(self):
        times2 = cuda.declare_device("times2", "int32(int32)", link=times2_cu)

        @cuda.jit
        def device_func(x):
            return times2(x)

        @cuda.jit
        def kernel(r, x):
            i = cuda.grid(1)
            if i < len(r):
                r[i] = device_func(x[i]) + device_func(x[i] + i)

        x = np.arange(10, dtype=np.int32)
        r = np.empty_like(x)

        kernel[1, 32](r, x)

        np.testing.assert_equal(r, x * 6)

    def test_call_declared_function_twice(self):
        times2 = cuda.declare_device("times2", "int32(int32)", link=times2_cu)

        @cuda.jit
        def kernel(r, x):
            i = cuda.grid(1)
            if i < len(r):
                r[i] = times2(x[i]) + times2(x[i] + i)

        x = np.arange(10, dtype=np.int32)
        r = np.empty_like(x)

        kernel[1, 32](r, x)

        np.testing.assert_equal(r, x * 6)

    def test_declared_in_called_function_and_parent(self):
        times2 = cuda.declare_device("times2", "int32(int32)", link=times2_cu)

        @cuda.jit
        def device_func(x):
            return times2(x)

        @cuda.jit
        def kernel(r, x):
            i = cuda.grid(1)
            if i < len(r):
                r[i] = device_func(x[i]) + times2(x[i])

        x = np.arange(10, dtype=np.int32)
        r = np.empty_like(x)

        kernel[1, 32](r, x)

        np.testing.assert_equal(r, x * 4)

    def test_call_two_different_declared_functions(self):
        times2 = cuda.declare_device("times2", "int32(int32)", link=times2_cu)
        times3 = cuda.declare_device("times3", "int32(int32)", link=times3_cu)

        @cuda.jit
        def kernel(r, x):
            i = cuda.grid(1)
            if i < len(r):
                r[i] = times2(x[i]) + times3(x[i])

        x = np.arange(10, dtype=np.int32)
        r = np.empty_like(x)

        kernel[1, 32](r, x)

        np.testing.assert_equal(r, x * 5)


if __name__ == "__main__":
    unittest.main()



from fnmatch import fnmatch
from numba.cuda.testing import ensure_supported_ccs_initialized
from numba.testing import unittest
from numba import cuda
from os.path import dirname, isfile, join, normpath, relpath, splitext

import os
import sys
import traceback


# Copied and modified from numba/testing/__init__.py, to handle the difference
# between the top dirs for Numba and the CUDA target
def load_testsuite(loader, dir):
    """Find tests in 'dir'."""
    top_level_dir = dirname(dirname(dirname(dirname(__file__))))
    try:
        suite = unittest.TestSuite()
        files = []
        for f in os.listdir(dir):
            path = join(dir, f)
            if isfile(path) and fnmatch(f, "test_*.py"):
                files.append(f)
            elif isfile(join(path, "__init__.py")):
                suite.addTests(
                    loader.discover(path, top_level_dir=top_level_dir)
                )
        for f in files:
            # turn 'f' into a filename relative to the toplevel dir and
            # translate it to a module name. This differs from the
            # implementation in Numba, because the toplevel dir is the
            # numba_cuda module location, not the numba one.
            f = relpath(join(dir, f), top_level_dir)
            f = splitext(normpath(f.replace(os.path.sep, ".")))[0]
            suite.addTests(loader.loadTestsFromName(f))
        return suite
    except Exception:
        traceback.print_exc(file=sys.stderr)
        sys.exit(-1)


def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    this_dir = dirname(__file__)
    ensure_supported_ccs_initialized()
    suite.addTests(load_testsuite(loader, join(this_dir, "nocuda")))
    if cuda.is_available():
        suite.addTests(load_testsuite(loader, join(this_dir, "cudasim")))
        gpus = cuda.list_devices()
        if gpus and gpus[0].compute_capability >= (2, 0):
            suite.addTests(load_testsuite(loader, join(this_dir, "cudadrv")))
            suite.addTests(load_testsuite(loader, join(this_dir, "cudapy")))
            suite.addTests(load_testsuite(loader, join(this_dir, "nrt")))
            suite.addTests(
                load_testsuite(loader, join(this_dir, "doc_examples"))
            )
        else:
            print("skipped CUDA tests because GPU CC < 2.0")
    else:
        print("skipped CUDA tests")
    return suite



import sys
import os
import multiprocessing as mp
import warnings

from numba.core.config import IS_WIN32, IS_OSX
from numba.core.errors import NumbaWarning
from numba.cuda.cudadrv import nvvm
from numba.cuda.testing import (
    unittest,
    skip_on_cudasim,
    SerialMixin,
    skip_unless_conda_cudatoolkit,
)
from numba.cuda.cuda_paths import (
    _get_libdevice_path_decision,
    _get_nvvm_path_decision,
    _get_cudalib_dir_path_decision,
    get_system_ctk,
)


has_cuda = nvvm.is_available()
has_mp_get_context = hasattr(mp, "get_context")


class LibraryLookupBase(SerialMixin, unittest.TestCase):
    def setUp(self):
        ctx = mp.get_context("spawn")

        qrecv = ctx.Queue()
        qsend = ctx.Queue()
        self.qsend = qsend
        self.qrecv = qrecv
        self.child_process = ctx.Process(
            target=check_lib_lookup,
            args=(qrecv, qsend),
            daemon=True,
        )
        self.child_process.start()

    def tearDown(self):
        self.qsend.put(self.do_terminate)
        self.child_process.join(3)
        # Ensure the process is terminated
        self.assertIsNotNone(self.child_process)

    def remote_do(self, action):
        self.qsend.put(action)
        out = self.qrecv.get()
        self.assertNotIsInstance(out, BaseException)
        return out

    @staticmethod
    def do_terminate():
        return False, None


def remove_env(name):
    try:
        del os.environ[name]
    except KeyError:
        return False
    else:
        return True


def check_lib_lookup(qout, qin):
    status = True
    while status:
        try:
            action = qin.get()
        except Exception as e:
            qout.put(e)
            status = False
        else:
            try:
                with warnings.catch_warnings(record=True) as w:
                    warnings.simplefilter("always", NumbaWarning)
                    status, result = action()
                qout.put(result + (w,))
            except Exception as e:
                qout.put(e)
                status = False


@skip_on_cudasim("Library detection unsupported in the simulator")
@unittest.skipUnless(has_mp_get_context, "mp.get_context not available")
@skip_unless_conda_cudatoolkit("test assumes conda installed cudatoolkit")
class TestLibDeviceLookUp(LibraryLookupBase):
    def test_libdevice_path_decision(self):
        # Check that the default is using conda environment
        by, info, warns = self.remote_do(self.do_clear_envs)
        if has_cuda:
            self.assertEqual(by, "Conda environment")
        else:
            self.assertEqual(by, "<unknown>")
            self.assertIsNone(info)
        self.assertFalse(warns)
        # Check that CUDA_HOME works by removing conda-env
        by, info, warns = self.remote_do(self.do_set_cuda_home)
        self.assertEqual(by, "CUDA_HOME")
        self.assertEqual(info, os.path.join("mycudahome", "nvvm", "libdevice"))
        self.assertFalse(warns)

        if get_system_ctk() is None:
            # Fake remove conda environment so no cudatoolkit is available
            by, info, warns = self.remote_do(self.do_clear_envs)
            self.assertEqual(by, "<unknown>")
            self.assertIsNone(info)
            self.assertFalse(warns)
        else:
            # Use system available cudatoolkit
            by, info, warns = self.remote_do(self.do_clear_envs)
            self.assertEqual(by, "System")
            self.assertFalse(warns)

    @staticmethod
    def do_clear_envs():
        remove_env("CUDA_HOME")
        remove_env("CUDA_PATH")
        return True, _get_libdevice_path_decision()

    @staticmethod
    def do_set_cuda_home():
        os.environ["CUDA_HOME"] = os.path.join("mycudahome")
        _fake_non_conda_env()
        return True, _get_libdevice_path_decision()


@skip_on_cudasim("Library detection unsupported in the simulator")
@unittest.skipUnless(has_mp_get_context, "mp.get_context not available")
@skip_unless_conda_cudatoolkit("test assumes conda installed cudatoolkit")
class TestNvvmLookUp(LibraryLookupBase):
    def test_nvvm_path_decision(self):
        # Check that the default is using conda environment
        by, info, warns = self.remote_do(self.do_clear_envs)
        if has_cuda:
            self.assertEqual(by, "Conda environment")
        else:
            self.assertEqual(by, "<unknown>")
            self.assertIsNone(info)
        self.assertFalse(warns)
        # Check that CUDA_HOME works by removing conda-env
        by, info, warns = self.remote_do(self.do_set_cuda_home)
        self.assertEqual(by, "CUDA_HOME")
        self.assertFalse(warns)
        if IS_WIN32:
            self.assertEqual(info, os.path.join("mycudahome", "nvvm", "bin"))
        elif IS_OSX:
            self.assertEqual(info, os.path.join("mycudahome", "nvvm", "lib"))
        else:
            self.assertEqual(info, os.path.join("mycudahome", "nvvm", "lib64"))

        if get_system_ctk() is None:
            # Fake remove conda environment so no cudatoolkit is available
            by, info, warns = self.remote_do(self.do_clear_envs)
            self.assertEqual(by, "<unknown>")
            self.assertIsNone(info)
            self.assertFalse(warns)
        else:
            # Use system available cudatoolkit
            by, info, warns = self.remote_do(self.do_clear_envs)
            self.assertEqual(by, "System")
            self.assertFalse(warns)

    @staticmethod
    def do_clear_envs():
        remove_env("CUDA_HOME")
        remove_env("CUDA_PATH")
        return True, _get_nvvm_path_decision()

    @staticmethod
    def do_set_cuda_home():
        os.environ["CUDA_HOME"] = os.path.join("mycudahome")
        _fake_non_conda_env()
        return True, _get_nvvm_path_decision()


@skip_on_cudasim("Library detection unsupported in the simulator")
@unittest.skipUnless(has_mp_get_context, "mp.get_context not available")
@skip_unless_conda_cudatoolkit("test assumes conda installed cudatoolkit")
class TestCudaLibLookUp(LibraryLookupBase):
    def test_cudalib_path_decision(self):
        # Check that the default is using conda environment
        by, info, warns = self.remote_do(self.do_clear_envs)
        if has_cuda:
            self.assertEqual(by, "Conda environment")
        else:
            self.assertEqual(by, "<unknown>")
            self.assertIsNone(info)
        self.assertFalse(warns)

        # Check that CUDA_HOME works by removing conda-env
        self.remote_do(self.do_clear_envs)
        by, info, warns = self.remote_do(self.do_set_cuda_home)
        self.assertEqual(by, "CUDA_HOME")
        self.assertFalse(warns)
        if IS_WIN32:
            self.assertEqual(info, os.path.join("mycudahome", "bin"))
        elif IS_OSX:
            self.assertEqual(info, os.path.join("mycudahome", "lib"))
        else:
            self.assertEqual(info, os.path.join("mycudahome", "lib64"))
        if get_system_ctk() is None:
            # Fake remove conda environment so no cudatoolkit is available
            by, info, warns = self.remote_do(self.do_clear_envs)
            self.assertEqual(by, "<unknown>")
            self.assertIsNone(info)
            self.assertFalse(warns)
        else:
            # Use system available cudatoolkit
            by, info, warns = self.remote_do(self.do_clear_envs)
            self.assertEqual(by, "System")
            self.assertFalse(warns)

    @staticmethod
    def do_clear_envs():
        remove_env("CUDA_HOME")
        remove_env("CUDA_PATH")
        return True, _get_cudalib_dir_path_decision()

    @staticmethod
    def do_set_cuda_home():
        os.environ["CUDA_HOME"] = os.path.join("mycudahome")
        _fake_non_conda_env()
        return True, _get_cudalib_dir_path_decision()


def _fake_non_conda_env():
    """
    Monkeypatch sys.prefix to hide the fact we are in a conda-env
    """
    sys.prefix = ""


if __name__ == "__main__":
    unittest.main()



"""CUDA Toolkit libraries lookup utilities.

CUDA Toolkit libraries can be available via either:

- the `cuda-nvcc` and `cuda-nvrtc` conda packages for CUDA 12,
- the `cudatoolkit` conda package for CUDA 11,
- a user supplied location from CUDA_HOME,
- a system wide location,
- package-specific locations (e.g. the Debian NVIDIA packages),
- or can be discovered by the system loader.
"""

import os
import sys
import ctypes

from numba.misc.findlib import find_lib
from numba.cuda.cuda_paths import get_cuda_paths
from numba.cuda.cudadrv.driver import locate_driver_and_loader, load_driver
from numba.cuda.cudadrv.error import CudaSupportError
from numba.core import config


if sys.platform == "win32":
    _dllnamepattern = "%s.dll"
    _staticnamepattern = "%s.lib"
elif sys.platform == "darwin":
    _dllnamepattern = "lib%s.dylib"
    _staticnamepattern = "lib%s.a"
else:
    _dllnamepattern = "lib%s.so"
    _staticnamepattern = "lib%s.a"


def get_libdevice():
    d = get_cuda_paths()
    paths = d["libdevice"].info
    return paths


def open_libdevice():
    with open(get_libdevice(), "rb") as bcfile:
        return bcfile.read()


def get_cudalib(lib, static=False):
    """
    Find the path of a CUDA library based on a search of known locations. If
    the search fails, return a generic filename for the library (e.g.
    'libnvvm.so' for 'nvvm') so that we may attempt to load it using the system
    loader's search mechanism.
    """
    if lib in {"nvrtc", "nvvm"}:
        return get_cuda_paths()[lib].info or _dllnamepattern % lib
    else:
        dir_type = "static_cudalib_dir" if static else "cudalib_dir"
        libdir = get_cuda_paths()[dir_type].info

    candidates = find_lib(lib, libdir, static=static)
    namepattern = _staticnamepattern if static else _dllnamepattern
    return max(candidates) if candidates else namepattern % lib


def get_cuda_include_dir():
    """
    Find the path to cuda include dir based on a list of default locations.
    Note that this does not list the `CUDA_INCLUDE_PATH` entry in user
    configuration.
    """

    return get_cuda_paths()["include_dir"].info


def check_cuda_include_dir(path):
    if path is None or not os.path.exists(path):
        raise FileNotFoundError(f"{path} not found")

    if not os.path.exists(os.path.join(path, "cuda_runtime.h")):
        raise FileNotFoundError(f"Unable to find cuda_runtime.h from {path}")


def open_cudalib(lib):
    path = get_cudalib(lib)
    return ctypes.CDLL(path)


def check_static_lib(path):
    if not os.path.isfile(path):
        raise FileNotFoundError(f"{path} not found")


def _get_source_variable(lib, static=False):
    if lib == "nvvm":
        return get_cuda_paths()["nvvm"].by
    elif lib == "nvrtc":
        return get_cuda_paths()["nvrtc"].by
    elif lib == "libdevice":
        return get_cuda_paths()["libdevice"].by
    elif lib == "include_dir":
        return get_cuda_paths()["include_dir"].by
    else:
        dir_type = "static_cudalib_dir" if static else "cudalib_dir"
        return get_cuda_paths()[dir_type].by


def test():
    """Test library lookup.  Path info is printed to stdout."""
    failed = False

    # Check for the driver
    try:
        dlloader, candidates = locate_driver_and_loader()
        print("Finding driver from candidates:")
        for location in candidates:
            print(f"\t{location}")
        print(f"Using loader {dlloader}")
        print("\tTrying to load driver", end="...")
        dll, path = load_driver(dlloader, candidates)
        print("\tok")
        print(f"\t\tLoaded from {path}")
    except CudaSupportError as e:
        print(f"\tERROR: failed to open driver: {e}")
        failed = True

    # Find the absolute location of the driver on Linux. Various driver-related
    # issues have been reported by WSL2 users, and it is almost always due to a
    # Linux (i.e. not- WSL2) driver being installed in a WSL2 system.
    # Providing the absolute location of the driver indicates its version
    # number in the soname (e.g. "libcuda.so.530.30.02"), which can be used to
    # look up whether the driver was intended for "native" Linux.
    if sys.platform == "linux" and not failed:
        pid = os.getpid()
        mapsfile = os.path.join(os.path.sep, "proc", f"{pid}", "maps")
        try:
            with open(mapsfile) as f:
                maps = f.read()
        # It's difficult to predict all that might go wrong reading the maps
        # file - in case various error conditions ensue (the file is not found,
        # not readable, etc.) we use OSError to hopefully catch any of them.
        except OSError:
            # It's helpful to report that this went wrong to the user, but we
            # don't set failed to True because this doesn't have any connection
            # to actual CUDA functionality.
            print(
                f"\tERROR: Could not open {mapsfile} to determine absolute "
                "path to libcuda.so"
            )
        else:
            # In this case we could read the maps, so we can report the
            # relevant ones to the user
            locations = set(s for s in maps.split() if "libcuda.so" in s)
            print("\tMapped libcuda.so paths:")
            for location in locations:
                print(f"\t\t{location}")

    # Checks for dynamic libraries
    libs = "nvvm nvrtc cudart".split()
    for lib in libs:
        path = get_cudalib(lib)
        print("Finding {} from {}".format(lib, _get_source_variable(lib)))
        print("\tLocated at", path)

        try:
            print("\tTrying to open library", end="...")
            open_cudalib(lib)
            print("\tok")
        except OSError as e:
            print("\tERROR: failed to open %s:\n%s" % (lib, e))
            failed = True

    # Check for cudadevrt (the only static library)
    lib = "cudadevrt"
    path = get_cudalib(lib, static=True)
    print(
        "Finding {} from {}".format(lib, _get_source_variable(lib, static=True))
    )
    print("\tLocated at", path)

    try:
        print("\tChecking library", end="...")
        check_static_lib(path)
        print("\tok")
    except FileNotFoundError as e:
        print("\tERROR: failed to find %s:\n%s" % (lib, e))
        failed = True

    # Check for libdevice
    where = _get_source_variable("libdevice")
    print(f"Finding libdevice from {where}")
    path = get_libdevice()
    print("\tLocated at", path)

    try:
        print("\tChecking library", end="...")
        check_static_lib(path)
        print("\tok")
    except FileNotFoundError as e:
        print("\tERROR: failed to find %s:\n%s" % (lib, e))
        failed = True

    # Check cuda include paths

    print("Include directory configuration variable:")
    print(f"\tCUDA_INCLUDE_PATH={config.CUDA_INCLUDE_PATH}")

    where = _get_source_variable("include_dir")
    print(f"Finding include directory from {where}")
    include = get_cuda_include_dir()
    print("\tLocated at", include)
    try:
        print("\tChecking include directory", end="...")
        check_cuda_include_dir(include)
        print("\tok")
    except FileNotFoundError as e:
        print("\tERROR: failed to find cuda include directory:\n%s" % e)
        failed = True

    return not failed



from numba.cuda.deviceufunc import GUFuncEngine
import unittest


def template(signature, shapes, expects):
    gufb = GUFuncEngine.from_signature(signature)
    sch = gufb.schedule(shapes)
    for k, v in expects.items():
        got = getattr(sch, k)
        if got != v:
            fmt = "error for %s: got=%s but expect=%s"
            raise AssertionError(fmt % (k, got, v))


class TestGUFuncScheduling(unittest.TestCase):
    def test_signature_1(self):
        signature = "(m, n), (n, p) -> (m, p)"
        shapes = (100, 4, 5), (1, 5, 7)
        expects = dict(
            ishapes=[(4, 5), (5, 7)],
            oshapes=[(4, 7)],
            loopdims=(100,),
            pinned=[False, True],
        )
        template(signature, shapes, expects)

    def test_signature_2(self):
        signature = "(m, n), (n, p) -> (m, p)"
        shapes = (100, 4, 5), (100, 5, 7)
        expects = dict(
            ishapes=[(4, 5), (5, 7)],
            oshapes=[(4, 7)],
            loopdims=(100,),
            pinned=[False, False],
        )
        template(signature, shapes, expects)

    def test_signature_3(self):
        signature = "(m, n), (n, p) -> (m, p)"
        shapes = (12, 34, 4, 5), (12, 34, 5, 7)
        expects = dict(
            ishapes=[(4, 5), (5, 7)],
            oshapes=[(4, 7)],
            loopdims=(12, 34),
            pinned=[False, False],
        )
        template(signature, shapes, expects)

    def test_signature_4(self):
        signature = "(m, n), (n, p) -> (m, p)"
        shapes = (4, 5), (5, 7)
        expects = dict(
            ishapes=[(4, 5), (5, 7)],
            oshapes=[(4, 7)],
            loopdims=(),
            pinned=[False, False],
        )
        template(signature, shapes, expects)

    def test_signature_5(self):
        signature = "(a), (a) -> (a)"
        shapes = (5,), (5,)
        expects = dict(
            ishapes=[(5,), (5,)],
            oshapes=[(5,)],
            loopdims=(),
            pinned=[False, False],
        )
        template(signature, shapes, expects)

    def test_signature_6(self):
        signature = "(), () -> ()"
        shapes = (5,), (5,)
        expects = dict(
            ishapes=[(), ()], oshapes=[()], loopdims=(5,), pinned=[False, False]
        )
        template(signature, shapes, expects)

    def test_signature_7(self):
        signature = "(), () -> ()"
        shapes = (5,), ()
        expects = dict(
            ishapes=[(), ()], oshapes=[()], loopdims=(5,), pinned=[False, True]
        )
        template(signature, shapes, expects)


if __name__ == "__main__":
    unittest.main()



from .decorators import jit
import numba


@jit(device=True)
def all_sync(mask, predicate):
    """
    If for all threads in the masked warp the predicate is true, then
    a non-zero value is returned, otherwise 0 is returned.
    """
    return numba.cuda.vote_sync_intrinsic(mask, 0, predicate)[1]


@jit(device=True)
def any_sync(mask, predicate):
    """
    If for any thread in the masked warp the predicate is true, then
    a non-zero value is returned, otherwise 0 is returned.
    """
    return numba.cuda.vote_sync_intrinsic(mask, 1, predicate)[1]


@jit(device=True)
def eq_sync(mask, predicate):
    """
    If for all threads in the masked warp the boolean predicate is the same,
    then a non-zero value is returned, otherwise 0 is returned.
    """
    return numba.cuda.vote_sync_intrinsic(mask, 2, predicate)[1]


@jit(device=True)
def ballot_sync(mask, predicate):
    """
    Returns a mask of all threads in the warp whose predicate is true,
    and are within the given mask.
    """
    return numba.cuda.vote_sync_intrinsic(mask, 3, predicate)[0]



import re
import itertools
import numpy as np
from numba import cuda
from numba.core.errors import TypingError
from numba.cuda.testing import CUDATestCase
import unittest


# Set to true if you want to see dots printed for each subtest.
NOISY = False


# In order to verify the alignment of the local and shared memory arrays, we
# inspect the LLVM IR of the generated kernel using the following regexes.

# Shared memory example:
# @"_cudapy_smem_38" = addrspace(3) global [1 x i8] undef, align 16
SMEM_PATTERN = re.compile(
    r'^@"_cudapy_smem_\d+".*?align (\d+)',
    re.MULTILINE,
)

# Local memory example:
# %"_cudapy_lmem" = alloca [1 x i8], align 64
LMEM_PATTERN = re.compile(
    r'^\s*%"_cudapy_lmem".*?align (\d+)',
    re.MULTILINE,
)


DTYPES = [np.uint8, np.uint32, np.uint64]

# Add in some record dtypes with and without alignment.
for align in (True, False):
    DTYPES += [
        np.dtype(
            [
                ("a", np.uint8),
                ("b", np.int32),
                ("c", np.float64),
            ],
            align=align,
        ),
        np.dtype(
            [
                ("a", np.uint32),
                ("b", np.uint8),
            ],
            align=align,
        ),
        np.dtype(
            [
                ("a", np.uint8),
                ("b", np.int32),
                ("c", np.float64),
                ("d", np.complex64),
                ("e", (np.uint8, 5)),
            ],
            align=align,
        ),
    ]

# N.B. We name the test class TestArrayAddressAlignment to avoid name conflict
#      with the test_alignment.TestArrayAlignment class.


class TestArrayAddressAlignment(CUDATestCase):
    """
    Test cuda.local.array and cuda.shared.array support for an alignment
    keyword argument.
    """

    def test_array_alignment_1d(self):
        shapes = (1, 8, 50)
        alignments = (None, 16, 64, 256)
        array_types = [(0, "local"), (1, "shared")]
        self._do_test(array_types, shapes, DTYPES, alignments)

    def test_array_alignment_2d(self):
        shapes = ((2, 3),)
        alignments = (None, 16, 64, 256)
        array_types = [(0, "local"), (1, "shared")]
        self._do_test(array_types, shapes, DTYPES, alignments)

    def test_array_alignment_3d(self):
        shapes = ((2, 3, 4), (1, 4, 5))
        alignments = (None, 16, 64, 256)
        array_types = [(0, "local"), (1, "shared")]
        self._do_test(array_types, shapes, DTYPES, alignments)

    def _do_test(self, array_types, shapes, dtypes, alignments):
        items = itertools.product(array_types, shapes, dtypes, alignments)

        for (which, array_type), shape, dtype, alignment in items:
            with self.subTest(
                array_type=array_type,
                shape=shape,
                dtype=dtype,
                alignment=alignment,
            ):

                @cuda.jit
                def f(loc, shrd, which):
                    i = cuda.grid(1)
                    if which == 0:
                        local_array = cuda.local.array(
                            shape=shape,
                            dtype=dtype,
                            alignment=alignment,
                        )
                        if i == 0:
                            loc[0] = local_array.ctypes.data
                    else:
                        shared_array = cuda.shared.array(
                            shape=shape,
                            dtype=dtype,
                            alignment=alignment,
                        )
                        if i == 0:
                            shrd[0] = shared_array.ctypes.data

                loc = np.zeros(1, dtype=np.uint64)
                shrd = np.zeros(1, dtype=np.uint64)
                f[1, 1](loc, shrd, which)

                kernel = f.overloads[f.signatures[0]]
                llvm_ir = kernel.inspect_llvm()

                if alignment is None:
                    if which == 0:
                        # Local memory shouldn't have any alignment information
                        # when no alignment is specified.
                        match = LMEM_PATTERN.findall(llvm_ir)
                        self.assertEqual(len(match), 0)
                    else:
                        # Shared memory should at least have a power-of-two
                        # alignment when no alignment is specified.
                        match = SMEM_PATTERN.findall(llvm_ir)
                        self.assertEqual(len(match), 1)

                        alignment = int(match[0])
                        # Verify alignment is a power of two.
                        self.assertTrue(alignment & (alignment - 1) == 0)
                else:
                    # Verify alignment is in the LLVM IR.
                    if which == 0:
                        match = LMEM_PATTERN.findall(llvm_ir)
                        self.assertEqual(len(match), 1)
                        actual_alignment = int(match[0])
                        self.assertEqual(alignment, actual_alignment)
                    else:
                        match = SMEM_PATTERN.findall(llvm_ir)
                        self.assertEqual(len(match), 1)
                        actual_alignment = int(match[0])
                        self.assertEqual(alignment, actual_alignment)

                    # Also verify that the address of the array is aligned.
                    # If this fails, there problem is likely with NVVM.
                    address = loc[0] if which == 0 else shrd[0]
                    alignment_mod = int(address % alignment)
                    self.assertEqual(alignment_mod, 0)

                if NOISY:
                    print(".", end="", flush=True)

    def test_invalid_aligments(self):
        shapes = (1, 50)
        dtypes = (np.uint8, np.uint64)
        invalid_alignment_values = (-1, 0, 3, 17, 33)
        invalid_alignment_types = ("1.0", "1", "foo", 1.0, 1.5, 3.2)
        alignments = invalid_alignment_values + invalid_alignment_types
        array_types = [(0, "local"), (1, "shared")]

        # Use regex pattern to match error message, handling potential ANSI
        # color codes which appear on CI.
        expected_invalid_type_error_regex = (
            r"RequireLiteralValue:.*alignment must be a constant integer"
        )

        items = itertools.product(array_types, shapes, dtypes, alignments)

        for (which, array_type), shape, dtype, alignment in items:
            with self.subTest(
                array_type=array_type,
                shape=shape,
                dtype=dtype,
                alignment=alignment,
            ):
                if which == 0:

                    @cuda.jit
                    def f(dest_array):
                        i = cuda.grid(1)
                        local_array = cuda.local.array(
                            shape=shape,
                            dtype=dtype,
                            alignment=alignment,
                        )
                        if i == 0:
                            dest_array[0] = local_array.ctypes.data
                else:

                    @cuda.jit
                    def f(dest_array):
                        i = cuda.grid(1)
                        shared_array = cuda.shared.array(
                            shape=shape,
                            dtype=dtype,
                            alignment=alignment,
                        )
                        if i == 0:
                            dest_array[0] = shared_array.ctypes.data

                array = np.zeros(1, dtype=np.uint64)

                # The type of error we expect differs between an invalid value
                # that is still an int, and an invalid type.
                if isinstance(alignment, int):
                    self.assertRaisesRegex(
                        ValueError, r"Alignment must be.*", f[1, 1], array
                    )
                else:
                    self.assertRaisesRegex(
                        TypingError,
                        expected_invalid_type_error_regex,
                        f[1, 1],
                        array,
                    )

                if NOISY:
                    print(".", end="", flush=True)


if __name__ == "__main__":
    unittest.main()
